import io
import json
import time
import uuid
import zipfile

import pandas
from docx import Document as Docs
from google.cloud import pubsub_v1, storage
from python_docx_replace import docx_replace

from constant import Constant
from models import Document, StatusEnum, db


def get_bucket():
    client = storage.Client.from_service_account_json("service_account.json")
    bucket = client.get_bucket("manipulation-bucket")
    return bucket


def get_result(id, no):
    bucket = get_bucket()

    if bucket.blob(f"results/{id}/result_{no}.docx").exists():
        return bucket.blob(f"results/{id}/result_{no}.docx")

    return None


def get_template(id):
    bucket = get_bucket()

    blobs = bucket.list_blobs(prefix=f"templates/{id}/")

    for blob in blobs:
        if blob.name.endswith(".docx"):
            return blob

    return None


def upload_template(file):
    template_id = uuid.uuid4()
    bucket = get_bucket()

    blob = bucket.blob(f"templates/{template_id}/{file.filename}")
    blob.upload_from_file(file, content_type=file.content_type)

    blob_url = f"https://storage.googleapis.com/{bucket.name}/{blob.name}"

    return template_id, blob_url


def save_result(template_blob, document_id):
    bucket = get_bucket()

    data_blob = bucket.blob(f"results/{document_id}/data.csv")
    data_binary = data_blob.download_as_bytes()
    data = pandas.read_csv(io.BytesIO(data_binary))

    for idx, row in data.iterrows():
        file = io.BytesIO(template_blob)
        doc = Docs(file)
        docx_replace(doc, **row.to_dict())
        file_stream = io.BytesIO()
        doc.save(file_stream)
        file_stream.seek(0)

        blob = bucket.blob(f"results/{document_id}/result_{idx+1}.docx")
        blob.upload_from_file(file_stream)

        time.sleep(5)  

    document = Document.query.get(document_id)
    if document:
        zip_buffer = io.BytesIO()
        with zipfile.ZipFile(zip_buffer, "w") as zip_file:
            for blob in bucket.list_blobs(prefix=f"results/{document_id}/"):
                if blob.name.endswith(".docx"):
                    docx_content = blob.download_as_bytes()
                    zip_file.writestr(blob.name.split("/")[-1], docx_content)
        zip_buffer.seek(0)
        zip_blob = bucket.blob(f"results/{document_id}/results.zip")
        zip_blob.upload_from_file(zip_buffer, content_type="application/zip")

        zip_url = zip_blob.public_url

        document.URL = zip_url
        document.status = StatusEnum.COMPLETED
        db.session.commit()

    data_blob.delete()


def publish_document(template_id, data_csv, document_id):
    bucket = get_bucket()

    blob = bucket.blob(f"results/{document_id}/data.csv")
    blob.upload_from_file(data_csv)

    publisher = pubsub_v1.PublisherClient.from_service_account_json(
        "service_account.json"
    )
    topic_path = publisher.topic_path("law-project-422612", "manipulation-topic")

    bytes_data = f"{template_id},{document_id}".encode()
    publisher.publish(topic_path, bytes_data)

    return document_id


def check_progress(document_id, socketio):
    bucket = get_bucket()

    data_blob = bucket.blob(f"results/{document_id}/data.csv")
    data_binary = data_blob.download_as_bytes()
    data = pandas.read_csv(io.BytesIO(data_binary))

    total_tasks = len(data)

    while bucket.blob(f"results/{document_id}/data.csv").exists():
        completed_tasks = sum(
            1 for _ in bucket.list_blobs(prefix=f"results/{document_id}")
        ) - 1
        percentage = (completed_tasks / total_tasks) * 100
        socketio.emit(
            f"progress-{document_id}", {"id": document_id, "percentage": percentage}
        )
