import base64
import csv
import io
import json
import base64
import os
from functools import wraps

import jwt
from dotenv import load_dotenv
from flask import Flask, g, jsonify, request, send_file
from flask_cors import CORS, cross_origin
from flask_migrate import Migrate
from flask_socketio import SocketIO, emit
from flask_sqlalchemy import SQLAlchemy

import helper
from constant import Constant
from models import Document, Template, db

app = Flask(__name__)
CORS(app, resources={r"/api/*": {"origins": "*"}})
app.config[
    "SQLALCHEMY_DATABASE_URI"
    # ] = "postgresql://postgres:postgres@localhost/law_docs"
] = "postgresql://postgres.cgvfgkyftgiqvpulcuug:MvkjEsYa6tpiEH3s@aws-0-ap-southeast-1.pooler.supabase.com:6543/postgres"

app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False
migrate = Migrate(app, db)

db.init_app(app)

socketio = SocketIO(app, cors_allowed_origins="*")

PREFIX = "/api/manipulate"


def validate_jwt_token(token):
    try:
        print("Token:",token)
        payload = jwt.decode(
            token,
            base64.b64decode(
                "b5883f7a7b671b16d4f90a64c1bf72f786f47b730d6c294d9d8618f21e5e70ca"
            ),
            algorithms=["HS256"],
        )
        return payload["id"]
    except jwt.ExpiredSignatureError as e:
        print(f"ExpiredSignatureError: {e}")
        return None
    except jwt.InvalidTokenError as e:
        print(f"InvalidTokenError: {e}")
        return None


def jwt_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        token = None
        auth_header = request.headers.get("Authorization")
        if auth_header and auth_header.startswith("Bearer "):
            token = auth_header.split(" ")[1]
        if not token:
            return jsonify({"message": "Token is missing!"}), 403
        user_id = validate_jwt_token(token)
        if user_id is None:
            return jsonify({"message": "Token is invalid!"}), 403
        g.user_id = user_id
        return f(*args, **kwargs)

    return decorated_function


@app.route("/", methods=["GET"])
def home():
    return {"data": "Welcome to the Document Generation service"}, 200


@app.route(PREFIX + "/templates", methods=["POST"])
@jwt_required
def save_templates():
    user_id = g.user_id

    file_template = request.files.get("file_template")
    if not file_template or file_template.mimetype not in Constant.MIME_TYPE_TEMPLATES:
        return {"message": "Template file is not valid"}, 400

    template_id, blob_url = helper.upload_template(file_template)

    new_template = Template(
        id=template_id,
        name=file_template.filename,
        URL=blob_url,
        user_id=user_id,
    )
    db.session.add(new_template)
    db.session.commit()

    data = {
        "template_id": template_id,
        "name": file_template.filename,
        "file_url": blob_url,
    }
    return data, 200


@app.route(PREFIX + "/templates", methods=["GET"])
@jwt_required
def list_templates():
    user_id = g.user_id

    templates = Template.query.filter_by(user_id=user_id).all()
    templates_list = [
        {
            "id": str(template.id),
            "name": template.name,
            "file_url": template.URL,
        }
        for template in templates
    ]
    return jsonify(templates_list), 200


@app.route(PREFIX + "/", methods=["GET"])
@jwt_required
def list_documents():
    user_id = g.user_id

    documents = Document.query.filter_by(user_id=user_id).all()
    documents_list = [
        {
            "id": str(doc.id),
            "name": doc.name,
            "blob_url": doc.URL,
            "status": doc.status.value,
            "created_at": doc.created_at.isoformat(),
        }
        for doc in documents
    ]
    return jsonify(documents_list), 200


@app.route(PREFIX + "/generate-result", methods=["POST"])
def generate_result():
    data_request = (
        request.get_json().get("message").get("data")
    )  # format data from Google Pub/Sub Publisher
    template_id, document_id = base64.b64decode(data_request).decode("utf-8").split(",")
    if not data_request:
        return {"message": "Data is not valid"}, 400

    template_blob = helper.get_template(template_id)
    if not template_blob:
        return {"message": "Template is not found"}, 404

    helper.save_result(template_blob.download_as_bytes(), document_id)

    data = {"document_id": document_id}
    return data, 200


@app.route(PREFIX + "/", methods=["POST", "OPTIONS"])
@jwt_required
def manipulate_docs():
    user_id = g.user_id

    is_bulk = request.args.get("bulk")
    is_async = request.args.get("async")
    destination = request.args.get("destination")

    file_template = request.files.get("template")
    if not file_template or file_template.mimetype not in Constant.MIME_TYPE_TEMPLATES:
        return {"message": "Template file is not valid"}, 400

    if is_bulk and is_async:
        data = request.files.get("data")
        if not data or data.mimetype != "text/csv":
            return {"message": "Data file is not valid"}, 400

        data_csv = data.stream

        template_id, blob_url = helper.upload_template(file_template)

        new_docs = Document(
            name="document-bulk",
            URL=blob_url,
            user_id=user_id,
        )
        db.session.add(new_docs)
        db.session.commit()
        document_id = helper.publish_document(template_id, data_csv, new_docs.id)

        # COBA base64 encode
        encoded_bytes = base64.b64encode(f"{template_id},{document_id}".encode("utf-8"))
        encoded_process = encoded_bytes.decode("utf-8")

        data = {
            "task_id": new_docs.id,
            "task_status": Constant.STATUS_IN_PROGRESS,
            "encoded_process": encoded_process,
        }
        return data, 202

    if is_async:
        # process async (publish data.csv Pub/Sub)
        request_data = request.form.get("data")
        try:
            request_data = json.loads(request_data)
        except Exception as e:
            print(e)
            return {"message": "Data is not valid"}, 400

        data_csv = io.StringIO()
        csv_writer = csv.writer(data_csv)
        csv_writer.writerow(request_data.get("keys"))
        csv_writer.writerows(request_data.get("values"))
        data_csv.seek(0)

        template_id, blob_url = helper.upload_template(file_template)
        new_docs = Document(
            name="document-not-bulk",
            URL=blob_url,
            user_id=user_id,
        )
        db.session.add(new_docs)
        db.session.commit()
        document_id = helper.publish_document(template_id, data_csv, new_docs.id)

        # COBA base64 encode
        encoded_bytes = base64.b64encode(f"{template_id},{document_id}".encode("utf-8"))
        encoded_process = encoded_bytes.decode("utf-8")

        data = {
            "task_id": new_docs.id,
            "task_status": Constant.STATUS_IN_PROGRESS,
            "encoded_process": encoded_process,
        }
        return data, 202

    # process synchronously (waiting for response?)
    template_id, blob_url = helper.upload_template(file_template)
    new_docs = Document(
        name=name,
        URL=blob_url,
        user_id=user_id,
    )
    db.session.add(new_docs)
    db.session.commit()
    document_id = helper.publish_document(template_id, data_csv, new_docs.id)

    data = {
        "task_id": new_docs.id,
        "task_status": "INPROGRESS",
    }
    return data, 200


@app.route(PREFIX + "/result", methods=["GET"])
@jwt_required
def check_status():
    document_id = request.args.get("document_id")

    socketio.start_background_task(
        target=helper.check_progress,
        document_id=document_id,
        socketio=socketio,
    )
    return {"document_id": document_id, "status": Constant.STATUS_IN_PROGRESS}, 202


if __name__ == "__main__":
    socketio.run(app, host="0.0.0.0", port=8000, debug=True)
